@@ -1,11 +1,12 @@
 """Type checking of attribute access"""
 
-from typing import cast, Callable, List, Optional, TypeVar
+from typing import cast, Callable, Optional, Union
+from typing_extensions import TYPE_CHECKING
 
 from mypy.types import (
     Type, Instance, AnyType, TupleType, TypedDictType, CallableType, FunctionLike, TypeVarDef,
-    Overloaded, TypeVarType, UnionType, PartialType, UninhabitedType, TypeOfAny, LiteralType,
-    DeletedType, NoneTyp, TypeType, function_type, get_type_vars,
+    Overloaded, TypeVarType, UnionType, PartialType, TypeOfAny, LiteralType,
+    DeletedType, NoneType, TypeType, function_type, get_type_vars, get_proper_type, ProperType
 )
 from mypy.nodes import (
     TypeInfo, FuncBase, Var, FuncDef, SymbolNode, Context, MypyFile, TypeVarExpr,
@@ -14,19 +15,19 @@ from mypy.nodes import (
 )
 from mypy.messages import MessageBuilder
 from mypy.maptype import map_instance_to_supertype
-from mypy.expandtype import expand_type_by_instance, expand_type, freshen_function_type_vars
+from mypy.expandtype import expand_type_by_instance, freshen_function_type_vars
 from mypy.erasetype import erase_typevars
-from mypy.infer import infer_type_arguments
-from mypy.typevars import fill_typevars
 from mypy.plugin import AttributeContext
 from mypy.typeanal import set_any_tvars
 from mypy import message_registry
 from mypy import subtypes
 from mypy import meet
-from mypy.typeops import tuple_fallback
+from mypy.typeops import (
+    tuple_fallback, bind_self, erase_to_bound, class_callable, type_object_type_from_function,
+    make_simplified_union,
+)
 
-MYPY = False
-if MYPY:  # import for forward declaration only
+if TYPE_CHECKING:  # import for forward declaration only
     import mypy.checker
 
 from mypy import state
@@ -45,11 +46,13 @@ class MemberContext:
                  original_type: Type,
                  context: Context,
                  msg: MessageBuilder,
-                 chk: 'mypy.checker.TypeChecker') -> None:
+                 chk: 'mypy.checker.TypeChecker',
+                 self_type: Optional[Type]) -> None:
         self.is_lvalue = is_lvalue
         self.is_super = is_super
         self.is_operator = is_operator
         self.original_type = original_type
+        self.self_type = self_type or original_type
         self.context = context  # Error context
         self.msg = msg
         self.chk = chk
@@ -60,9 +63,16 @@ class MemberContext:
     def not_ready_callback(self, name: str, context: Context) -> None:
         self.chk.handle_cannot_determine_type(name, context)
 
-    def copy_modified(self, messages: MessageBuilder) -> 'MemberContext':
-        return MemberContext(self.is_lvalue, self.is_super, self.is_operator,
-                             self.original_type, self.context, messages, self.chk)
+    def copy_modified(self, *, messages: Optional[MessageBuilder] = None,
+                      self_type: Optional[Type] = None) -> 'MemberContext':
+        mx = MemberContext(self.is_lvalue, self.is_super, self.is_operator,
+                           self.original_type, self.context, self.msg, self.chk,
+                           self.self_type)
+        if messages is not None:
+            mx.msg = messages
+        if self_type is not None:
+            mx.self_type = self_type
+        return mx
 
 
 def analyze_member_access(name: str,
@@ -75,7 +85,8 @@ def analyze_member_access(name: str,
                           original_type: Type,
                           chk: 'mypy.checker.TypeChecker',
                           override_info: Optional[TypeInfo] = None,
-                          in_literal_context: bool = False) -> Type:
+                          in_literal_context: bool = False,
+                          self_type: Optional[Type] = None) -> Type:
     """Return the type of attribute 'name' of 'typ'.
 
     The actual implementation is in '_analyze_member_access' and this docstring
@@ -91,7 +102,9 @@ def analyze_member_access(name: str,
     that we have available. When looking for an attribute of 'typ', we may perform
     recursive calls targeting the fallback type, and 'typ' may become some supertype
     of 'original_type'. 'original_type' is always preserved as the 'typ' type used in
-    the initial, non-recursive call.
+    the initial, non-recursive call. The 'self_type' is a component of 'original_type'
+    to which generic self should be bound (a narrower type that has a fallback to instance).
+    Currently this is used only for union types.
     """
     mx = MemberContext(is_lvalue,
                        is_super,
@@ -99,10 +112,13 @@ def analyze_member_access(name: str,
                        original_type,
                        context,
                        msg,
-                       chk=chk)
+                       chk=chk,
+                       self_type=self_type)
     result = _analyze_member_access(name, typ, mx, override_info)
-    if in_literal_context and isinstance(result, Instance) and result.final_value is not None:
-        return result.final_value
+    possible_literal = get_proper_type(result)
+    if (in_literal_context and isinstance(possible_literal, Instance) and
+            possible_literal.last_known_value is not None):
+        return possible_literal.last_known_value
     else:
         return result
 
@@ -113,6 +129,7 @@ def _analyze_member_access(name: str,
                            override_info: Optional[TypeInfo] = None) -> Type:
     # TODO: This and following functions share some logic with subtypes.find_member;
     #       consider refactoring.
+    typ = get_proper_type(typ)
     if isinstance(typ, Instance):
         return analyze_instance_member_access(name, typ, mx, override_info)
     elif isinstance(typ, AnyType):
@@ -123,17 +140,17 @@ def _analyze_member_access(name: str,
     elif isinstance(typ, FunctionLike) and typ.is_type_obj():
         return analyze_type_callable_member_access(name, typ, mx)
     elif isinstance(typ, TypeType):
-        return analyze_type_type_member_access(name, typ, mx)
+        return analyze_type_type_member_access(name, typ, mx, override_info)
     elif isinstance(typ, TupleType):
         # Actually look up from the fallback instance type.
-        return _analyze_member_access(name, tuple_fallback(typ), mx)
+        return _analyze_member_access(name, tuple_fallback(typ), mx, override_info)
     elif isinstance(typ, (TypedDictType, LiteralType, FunctionLike)):
         # Actually look up from the fallback instance type.
-        return _analyze_member_access(name, typ.fallback, mx)
-    elif isinstance(typ, NoneTyp):
+        return _analyze_member_access(name, typ.fallback, mx, override_info)
+    elif isinstance(typ, NoneType):
         return analyze_none_member_access(name, typ, mx)
     elif isinstance(typ, TypeVarType):
-        return _analyze_member_access(name, typ.upper_bound, mx)
+        return _analyze_member_access(name, typ.upper_bound, mx, override_info)
     elif isinstance(typ, DeletedType):
         mx.msg.deleted_as_rvalue(typ, mx.context)
         return AnyType(TypeOfAny.from_error)
@@ -183,7 +200,7 @@ def analyze_instance_member_access(name: str,
             # the first argument.
             pass
         else:
-            signature = bind_self(signature, mx.original_type)
+            signature = bind_self(signature, mx.self_type)
         typ = map_instance_to_supertype(typ, method.info)
         member_type = expand_type_by_instance(signature, typ)
         freeze_type_vars(member_type)
@@ -199,6 +216,7 @@ def analyze_type_callable_member_access(name: str,
     # Class attribute.
     # TODO super?
     ret_type = typ.items()[0].ret_type
+    assert isinstance(ret_type, ProperType)
     if isinstance(ret_type, TupleType):
         ret_type = tuple_fallback(ret_type)
     if isinstance(ret_type, Instance):
@@ -225,7 +243,10 @@ def analyze_type_callable_member_access(name: str,
         assert False, 'Unexpected type {}'.format(repr(ret_type))
 
 
-def analyze_type_type_member_access(name: str, typ: TypeType, mx: MemberContext) -> Type:
+def analyze_type_type_member_access(name: str,
+                                    typ: TypeType,
+                                    mx: MemberContext,
+                                    override_info: Optional[TypeInfo]) -> Type:
     # Similar to analyze_type_callable_attribute_access.
     item = None
     fallback = mx.builtin_type('builtins.type')
@@ -235,10 +256,13 @@ def analyze_type_type_member_access(name: str, typ: TypeType, mx: MemberContext)
         item = typ.item
     elif isinstance(typ.item, AnyType):
         mx = mx.copy_modified(messages=ignore_messages)
-        return _analyze_member_access(name, fallback, mx)
+        return _analyze_member_access(name, fallback, mx, override_info)
     elif isinstance(typ.item, TypeVarType):
-        if isinstance(typ.item.upper_bound, Instance):
-            item = typ.item.upper_bound
+        upper_bound = get_proper_type(typ.item.upper_bound)
+        if isinstance(upper_bound, Instance):
+            item = upper_bound
+        elif isinstance(upper_bound, TupleType):
+            item = tuple_fallback(upper_bound)
     elif isinstance(typ.item, TupleType):
         item = tuple_fallback(typ.item)
     elif isinstance(typ.item, FunctionLike) and typ.item.is_type_obj():
@@ -249,27 +273,30 @@ def analyze_type_type_member_access(name: str, typ: TypeType, mx: MemberContext)
             item = typ.item.item.type.metaclass_type
     if item and not mx.is_operator:
         # See comment above for why operators are skipped
-        result = analyze_class_attribute_access(item, name, mx)
+        result = analyze_class_attribute_access(item, name, mx, override_info)
         if result:
-            if not (isinstance(result, AnyType) and item.type.fallback_to_any):
+            if not (isinstance(get_proper_type(result), AnyType) and item.type.fallback_to_any):
                 return result
             else:
                 # We don't want errors on metaclass lookup for classes with Any fallback
                 mx = mx.copy_modified(messages=ignore_messages)
     if item is not None:
         fallback = item.type.metaclass_type or fallback
-    return _analyze_member_access(name, fallback, mx)
+    return _analyze_member_access(name, fallback, mx, override_info)
 
 
 def analyze_union_member_access(name: str, typ: UnionType, mx: MemberContext) -> Type:
     mx.msg.disable_type_names += 1
-    results = [_analyze_member_access(name, subtype, mx)
-               for subtype in typ.relevant_items()]
+    results = []
+    for subtype in typ.relevant_items():
+        # Self types should be bound to every individual item of a union.
+        item_mx = mx.copy_modified(self_type=subtype)
+        results.append(_analyze_member_access(name, subtype, item_mx))
     mx.msg.disable_type_names -= 1
-    return UnionType.make_simplified_union(results)
+    return make_simplified_union(results)
 
 
-def analyze_none_member_access(name: str, typ: NoneTyp, mx: MemberContext) -> Type:
+def analyze_none_member_access(name: str, typ: NoneType, mx: MemberContext) -> Type:
     if mx.chk.should_suppress_optional_error([typ]):
         return AnyType(TypeOfAny.from_error)
     is_python_3 = mx.chk.options.python_version[0] >= 3
@@ -310,7 +337,7 @@ def analyze_member_var_access(name: str,
         v = Var(name, type=type_object_type(vv, mx.builtin_type))
         v.info = info
 
-    if isinstance(vv, TypeAlias) and isinstance(vv.target, Instance):
+    if isinstance(vv, TypeAlias) and isinstance(get_proper_type(vv.target), Instance):
         # Similar to the above TypeInfo case, we allow using
         # qualified type aliases in runtime context if it refers to an
         # instance type. For example:
@@ -332,7 +359,8 @@ def analyze_member_var_access(name: str,
         return analyze_var(name, v, itype, info, mx, implicit=implicit)
     elif isinstance(v, FuncDef):
         assert False, "Did not expect a function"
-    elif not v and name not in ['__getattr__', '__setattr__', '__getattribute__']:
+    elif (not v and name not in ['__getattr__', '__setattr__', '__getattribute__'] and
+          not mx.is_operator):
         if not mx.is_lvalue:
             for method_name in ('__getattribute__', '__getattr__'):
                 method = info.get_method(method_name)
@@ -341,7 +369,7 @@ def analyze_member_var_access(name: str,
                 # that the attribute exists
                 if method and method.info.fullname() != 'builtins.object':
                     function = function_type(method, mx.builtin_type('builtins.function'))
-                    bound_method = bind_self(function, mx.original_type)
+                    bound_method = bind_self(function, mx.self_type)
                     typ = map_instance_to_supertype(itype, method.info)
                     getattr_type = expand_type_by_instance(bound_method, typ)
                     if isinstance(getattr_type, CallableType):
@@ -351,14 +379,14 @@ def analyze_member_var_access(name: str,
                         fullname = '{}.{}'.format(method.info.fullname(), name)
                         hook = mx.chk.plugin.get_attribute_hook(fullname)
                         if hook:
-                            result = hook(AttributeContext(mx.original_type, result,
-                                                           mx.context, mx.chk))
+                            result = hook(AttributeContext(get_proper_type(mx.original_type),
+                                                           result, mx.context, mx.chk))
                         return result
         else:
             setattr_meth = info.get_method('__setattr__')
             if setattr_meth and setattr_meth.info.fullname() != 'builtins.object':
                 setattr_func = function_type(setattr_meth, mx.builtin_type('builtins.function'))
-                bound_type = bind_self(setattr_func, mx.original_type)
+                bound_type = bind_self(setattr_func, mx.self_type)
                 typ = map_instance_to_supertype(itype, setattr_meth.info)
                 setattr_type = expand_type_by_instance(bound_type, typ)
                 if isinstance(setattr_type, CallableType) and len(setattr_type.arg_types) > 0:
@@ -403,9 +431,12 @@ def analyze_descriptor_access(instance_type: Type,
     Return:
         The return type of the appropriate ``__get__`` overload for the descriptor.
     """
+    instance_type = get_proper_type(instance_type)
+    descriptor_type = get_proper_type(descriptor_type)
+
     if isinstance(descriptor_type, UnionType):
         # Map the access over union types
-        return UnionType.make_simplified_union([
+        return make_simplified_union([
             analyze_descriptor_access(instance_type, typ, builtin_type,
                                       msg, context, chk=chk)
             for typ in descriptor_type.items
@@ -429,18 +460,20 @@ def analyze_descriptor_access(instance_type: Type,
 
     if isinstance(instance_type, FunctionLike) and instance_type.is_type_obj():
         owner_type = instance_type.items()[0].ret_type
-        instance_type = NoneTyp()
+        instance_type = NoneType()
     elif isinstance(instance_type, TypeType):
         owner_type = instance_type.item
-        instance_type = NoneTyp()
+        instance_type = NoneType()
     else:
         owner_type = instance_type
 
     _, inferred_dunder_get_type = chk.expr_checker.check_call(
         dunder_get_type,
-        [TempNode(instance_type), TempNode(TypeType.make_normalized(owner_type))],
+        [TempNode(instance_type, context=context),
+         TempNode(TypeType.make_normalized(owner_type), context=context)],
         [ARG_POS, ARG_POS], context)
 
+    inferred_dunder_get_type = get_proper_type(inferred_dunder_get_type)
     if isinstance(inferred_dunder_get_type, AnyType):
         # check_call failed, and will have reported an error
         return inferred_dunder_get_type
@@ -458,8 +491,9 @@ def instance_alias_type(alias: TypeAlias,
 
     As usual, we first erase any unbound type variables to Any.
     """
-    assert isinstance(alias.target, Instance), "Must be called only with aliases to classes"
-    target = set_any_tvars(alias.target, alias.alias_tvars, alias.line, alias.column)
+    target = get_proper_type(alias.target)
+    assert isinstance(target, Instance), "Must be called only with aliases to classes"
+    target = set_any_tvars(target, alias.alias_tvars, alias.line, alias.column)
     assert isinstance(target, Instance)
     tp = type_object_type(target.type, builtin_type)
     return expand_type_by_instance(tp, target)
@@ -491,7 +525,7 @@ def analyze_var(name: str,
             mx.msg.read_only_property(name, itype.type, mx.context)
         if mx.is_lvalue and var.is_classvar:
             mx.msg.cant_assign_to_classvar(name, mx.context)
-        result = t
+        result = t  # type: Type
         if var.is_initialized_in_class and isinstance(t, FunctionLike) and not t.is_type_obj():
             if mx.is_lvalue:
                 if var.is_property:
@@ -514,7 +548,7 @@ def analyze_var(name: str,
                 dispatched_type = meet.meet_types(mx.original_type, itype)
                 check_self_arg(functype, dispatched_type, var.is_classmethod, mx.context, name,
                                mx.msg)
-                signature = bind_self(functype, mx.original_type, var.is_classmethod)
+                signature = bind_self(functype, mx.self_type, var.is_classmethod)
                 if var.is_property:
                     # A property cannot have an overloaded type => the cast is fine.
                     assert isinstance(signature, CallableType)
@@ -532,11 +566,12 @@ def analyze_var(name: str,
         result = analyze_descriptor_access(mx.original_type, result, mx.builtin_type,
                                            mx.msg, mx.context, chk=mx.chk)
     if hook:
-        result = hook(AttributeContext(mx.original_type, result, mx.context, mx.chk))
+        result = hook(AttributeContext(get_proper_type(mx.original_type),
+                                       result, mx.context, mx.chk))
     return result
 
 
-def freeze_type_vars(member_type: Type) -> None:
+def freeze_type_vars(member_type: ProperType) -> None:
     if isinstance(member_type, CallableType):
         for v in member_type.variables:
             v.id.meta_level = 0
@@ -586,11 +621,16 @@ def check_self_arg(functype: FunctionLike,
 
 def analyze_class_attribute_access(itype: Instance,
                                    name: str,
-                                   mx: MemberContext) -> Optional[Type]:
+                                   mx: MemberContext,
+                                   override_info: Optional[TypeInfo] = None) -> Optional[Type]:
     """original_type is the type of E in the expression E.var"""
-    node = itype.type.get(name)
+    info = itype.type
+    if override_info:
+        info = override_info
+
+    node = info.get(name)
     if not node:
-        if itype.type.fallback_to_any:
+        if info.fallback_to_any:
             return AnyType(TypeOfAny.special_form)
         return None
 
@@ -611,10 +651,11 @@ def analyze_class_attribute_access(itype: Instance,
     # An assignment to final attribute on class object is also always an error,
     # independently of types.
     if mx.is_lvalue and not mx.chk.get_final_context():
-        check_final_member(name, itype.type, mx.msg, mx.context)
+        check_final_member(name, info, mx.msg, mx.context)
 
-    if itype.type.is_enum and not (mx.is_lvalue or is_decorated or is_method):
-        return itype
+    if info.is_enum and not (mx.is_lvalue or is_decorated or is_method):
+        enum_literal = LiteralType(name, fallback=itype)
+        return itype.copy_modified(last_known_value=enum_literal)
 
     t = node.type
     if t:
@@ -650,7 +691,7 @@ def analyze_class_attribute_access(itype: Instance,
             #     C[int].x  # Also an error, since C[int] is same as C at runtime
             if isinstance(t, TypeVarType) or get_type_vars(t):
                 # Exception: access on Type[...], including first argument of class methods is OK.
-                if not isinstance(mx.original_type, TypeType):
+                if not isinstance(get_proper_type(mx.original_type), TypeType):
                     mx.msg.fail(message_registry.GENERIC_INSTANCE_VAR_CLASS_ACCESS, mx.context)
 
             # Erase non-mapped variables, but keep mapped ones, even if there is an error.
@@ -661,8 +702,8 @@ def analyze_class_attribute_access(itype: Instance,
 
         is_classmethod = ((is_decorated and cast(Decorator, node.node).func.is_class)
                           or (isinstance(node.node, FuncBase) and node.node.is_class))
-        result = add_class_tvars(t, itype, isuper, is_classmethod, mx.builtin_type,
-                                 mx.original_type)
+        result = add_class_tvars(get_proper_type(t), itype, isuper, is_classmethod,
+                                 mx.builtin_type, mx.original_type)
         if not mx.is_lvalue:
             result = analyze_descriptor_access(mx.original_type, result, mx.builtin_type,
                                                mx.msg, mx.context, chk=mx.chk)
@@ -673,7 +714,7 @@ def analyze_class_attribute_access(itype: Instance,
 
     if isinstance(node.node, TypeVarExpr):
         mx.msg.fail(message_registry.CANNOT_USE_TYPEVAR_AS_EXPRESSION.format(
-                    itype.type.name(), name), mx.context)
+                    info.name(), name), mx.context)
         return AnyType(TypeOfAny.from_error)
 
     if isinstance(node.node, TypeInfo):
@@ -683,7 +724,8 @@ def analyze_class_attribute_access(itype: Instance,
         # Reference to a module object.
         return mx.builtin_type('types.ModuleType')
 
-    if isinstance(node.node, TypeAlias) and isinstance(node.node.target, Instance):
+    if (isinstance(node.node, TypeAlias) and
+            isinstance(get_proper_type(node.node.target), Instance)):
         return instance_alias_type(node.node, mx.builtin_type)
 
     if is_decorated:
@@ -697,7 +739,8 @@ def analyze_class_attribute_access(itype: Instance,
         return function_type(cast(FuncBase, node.node), mx.builtin_type('builtins.function'))
 
 
-def add_class_tvars(t: Type, itype: Instance, isuper: Optional[Instance], is_classmethod: bool,
+def add_class_tvars(t: ProperType, itype: Instance, isuper: Optional[Instance],
+                    is_classmethod: bool,
                     builtin_type: Callable[[str], Instance],
                     original_type: Type) -> Type:
     """Instantiate type variables during analyze_class_attribute_access,
@@ -750,7 +793,7 @@ def add_class_tvars(t: Type, itype: Instance, isuper: Optional[Instance], is_cla
     return t
 
 
-def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) -> Type:
+def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) -> ProperType:
     """Return the type of a type object.
 
     For a generic type G with type variables T and S the type is generally of form
@@ -763,25 +806,34 @@ def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) ->
 
     # We take the type from whichever of __init__ and __new__ is first
     # in the MRO, preferring __init__ if there is a tie.
-    init_method = info.get_method('__init__')
-    new_method = info.get_method('__new__')
-    if not init_method:
+    init_method = info.get('__init__')
+    new_method = info.get('__new__')
+    if not init_method or not is_valid_constructor(init_method.node):
         # Must be an invalid class definition.
         return AnyType(TypeOfAny.from_error)
     # There *should* always be a __new__ method except the test stubs
     # lack it, so just copy init_method in that situation
     new_method = new_method or init_method
+    if not is_valid_constructor(new_method.node):
+        # Must be an invalid class definition.
+        return AnyType(TypeOfAny.from_error)
 
-    init_index = info.mro.index(init_method.info)
-    new_index = info.mro.index(new_method.info)
+    # The two is_valid_constructor() checks ensure this.
+    assert isinstance(new_method.node, (SYMBOL_FUNCBASE_TYPES, Decorator))
+    assert isinstance(init_method.node, (SYMBOL_FUNCBASE_TYPES, Decorator))
+
+    init_index = info.mro.index(init_method.node.info)
+    new_index = info.mro.index(new_method.node.info)
 
     fallback = info.metaclass_type or builtin_type('builtins.type')
     if init_index < new_index:
-        method = init_method
+        method = init_method.node  # type: Union[FuncBase, Decorator]
+        is_new = False
     elif init_index > new_index:
-        method = new_method
+        method = new_method.node
+        is_new = True
     else:
-        if init_method.info.fullname() == 'builtins.object':
+        if init_method.node.info.fullname() == 'builtins.object':
             # Both are defined by object.  But if we've got a bogus
             # base class, we can't know for sure, so check for that.
             if info.fallback_to_any:
@@ -792,177 +844,32 @@ def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) ->
                                    arg_names=["_args", "_kwds"],
                                    ret_type=any_type,
                                    fallback=builtin_type('builtins.function'))
-                return class_callable(sig, info, fallback, None)
+                return class_callable(sig, info, fallback, None, is_new=False)
 
         # Otherwise prefer __init__ in a tie. It isn't clear that this
         # is the right thing, but __new__ caused problems with
         # typeshed (#5647).
-        method = init_method
+        method = init_method.node
+        is_new = False
     # Construct callable type based on signature of __init__. Adjust
     # return type and insert type arguments.
-    return type_object_type_from_function(method, info, fallback)
-
-
-def type_object_type_from_function(init_or_new: FuncBase,
-                                   info: TypeInfo,
-                                   fallback: Instance) -> FunctionLike:
-    signature = bind_self(function_type(init_or_new, fallback))
-
-    # The __init__ method might come from a generic superclass
-    # (init_or_new.info) with type variables that do not map
-    # identically to the type variables of the class being constructed
-    # (info). For example
-    #
-    #   class A(Generic[T]): def __init__(self, x: T) -> None: pass
-    #   class B(A[List[T]], Generic[T]): pass
-    #
-    # We need to first map B's __init__ to the type (List[T]) -> None.
-    signature = cast(FunctionLike,
-                     map_type_from_supertype(signature, info, init_or_new.info))
-    special_sig = None  # type: Optional[str]
-    if init_or_new.info.fullname() == 'builtins.dict':
-        # Special signature!
-        special_sig = 'dict'
-
-    if isinstance(signature, CallableType):
-        return class_callable(signature, info, fallback, special_sig)
+    if isinstance(method, FuncBase):
+        t = function_type(method, fallback)
     else:
-        # Overloaded __init__/__new__.
-        assert isinstance(signature, Overloaded)
-        items = []  # type: List[CallableType]
-        for item in signature.items():
-            items.append(class_callable(item, info, fallback, special_sig))
-        return Overloaded(items)
+        assert isinstance(method.type, ProperType)
+        assert isinstance(method.type, FunctionLike)  # is_valid_constructor() ensures this
+        t = method.type
+    return type_object_type_from_function(t, info, method.info, fallback, is_new)
 
 
-def class_callable(init_type: CallableType, info: TypeInfo, type_type: Instance,
-                   special_sig: Optional[str]) -> CallableType:
-    """Create a type object type based on the signature of __init__."""
-    variables = []  # type: List[TypeVarDef]
-    variables.extend(info.defn.type_vars)
-    variables.extend(init_type.variables)
+def is_valid_constructor(n: Optional[SymbolNode]) -> bool:
+    """Does this node represents a valid constructor method?
 
-    callable_type = init_type.copy_modified(
-        ret_type=fill_typevars(info), fallback=type_type, name=None, variables=variables,
-        special_sig=special_sig)
-    c = callable_type.with_name(info.name())
-    return c
-
-
-def map_type_from_supertype(typ: Type,
-                            sub_info: TypeInfo,
-                            super_info: TypeInfo) -> Type:
-    """Map type variables in a type defined in a supertype context to be valid
-    in the subtype context. Assume that the result is unique; if more than
-    one type is possible, return one of the alternatives.
-
-    For example, assume
-
-    . class D(Generic[S]) ...
-    . class C(D[E[T]], Generic[T]) ...
-
-    Now S in the context of D would be mapped to E[T] in the context of C.
+    This includes normal functions, overloaded functions, and decorators
+    that return a callable type.
     """
-    # Create the type of self in subtype, of form t[a1, ...].
-    inst_type = fill_typevars(sub_info)
-    if isinstance(inst_type, TupleType):
-        inst_type = tuple_fallback(inst_type)
-    # Map the type of self to supertype. This gets us a description of the
-    # supertype type variables in terms of subtype variables, i.e. t[t1, ...]
-    # so that any type variables in tN are to be interpreted in subtype
-    # context.
-    inst_type = map_instance_to_supertype(inst_type, super_info)
-    # Finally expand the type variables in type with those in the previously
-    # constructed type. Note that both type and inst_type may have type
-    # variables, but in type they are interpreted in supertype context while
-    # in inst_type they are interpreted in subtype context. This works even if
-    # the names of type variables in supertype and subtype overlap.
-    return expand_type_by_instance(typ, inst_type)
-
-
-F = TypeVar('F', bound=FunctionLike)
-
-
-def bind_self(method: F, original_type: Optional[Type] = None, is_classmethod: bool = False) -> F:
-    """Return a copy of `method`, with the type of its first parameter (usually
-    self or cls) bound to original_type.
-
-    If the type of `self` is a generic type (T, or Type[T] for classmethods),
-    instantiate every occurrence of type with original_type in the rest of the
-    signature and in the return type.
-
-    original_type is the type of E in the expression E.copy(). It is None in
-    compatibility checks. In this case we treat it as the erasure of the
-    declared type of self.
-
-    This way we can express "the type of self". For example:
-
-    T = TypeVar('T', bound='A')
-    class A:
-        def copy(self: T) -> T: ...
-
-    class B(A): pass
-
-    b = B().copy()  # type: B
-
-    """
-    if isinstance(method, Overloaded):
-        return cast(F, Overloaded([bind_self(c, original_type) for c in method.items()]))
-    assert isinstance(method, CallableType)
-    func = method
-    if not func.arg_types:
-        # invalid method. return something
-        return cast(F, func)
-    if func.arg_kinds[0] == ARG_STAR:
-        # The signature is of the form 'def foo(*args, ...)'.
-        # In this case we shouldn't drop the first arg,
-        # since func will be absorbed by the *args.
-
-        # TODO: infer bounds on the type of *args?
-        return cast(F, func)
-    self_param_type = func.arg_types[0]
-    if func.variables and (isinstance(self_param_type, TypeVarType) or
-                           (isinstance(self_param_type, TypeType) and
-                            isinstance(self_param_type.item, TypeVarType))):
-        if original_type is None:
-            # Type check method override
-            # XXX value restriction as union?
-            original_type = erase_to_bound(self_param_type)
-
-        ids = [x.id for x in func.variables]
-        typearg = infer_type_arguments(ids, self_param_type, original_type)[0]
-        if (is_classmethod and isinstance(typearg, UninhabitedType)
-                and isinstance(original_type, (Instance, TypeVarType, TupleType))):
-            # In case we call a classmethod through an instance x, fallback to type(x)
-            # TODO: handle Union
-            typearg = infer_type_arguments(ids, self_param_type, TypeType(original_type))[0]
-
-        def expand(target: Type) -> Type:
-            assert typearg is not None
-            return expand_type(target, {func.variables[0].id: typearg})
-
-        arg_types = [expand(x) for x in func.arg_types[1:]]
-        ret_type = expand(func.ret_type)
-        variables = func.variables[1:]
-    else:
-        arg_types = func.arg_types[1:]
-        ret_type = func.ret_type
-        variables = func.variables
-    if isinstance(original_type, CallableType) and original_type.is_type_obj():
-        original_type = TypeType.make_normalized(original_type.ret_type)
-    res = func.copy_modified(arg_types=arg_types,
-                             arg_kinds=func.arg_kinds[1:],
-                             arg_names=func.arg_names[1:],
-                             variables=variables,
-                             ret_type=ret_type,
-                             bound_args=[original_type])
-    return cast(F, res)
-
-
-def erase_to_bound(t: Type) -> Type:
-    if isinstance(t, TypeVarType):
-        return t.upper_bound
-    if isinstance(t, TypeType):
-        if isinstance(t.item, TypeVarType):
-            return TypeType.make_normalized(t.item.upper_bound)
-    return t
+    if isinstance(n, FuncBase):
+        return True
+    if isinstance(n, Decorator):
+        return isinstance(get_proper_type(n.type), FunctionLike)
+    return False
