@@ -6,14 +6,11 @@ This should be pretty lightweight and not depend on other mypy code (other than
 import json
 
 from typing import Any
+from typing_extensions import Final
 
 from mypy.ipc import IPCBase
 
-MYPY = False
-if MYPY:
-    from typing_extensions import Final
-
-STATUS_FILE = '.dmypy.json'  # type: Final
+DEFAULT_STATUS_FILE = '.dmypy.json'  # type: Final
 
 
 def receive(connection: IPCBase) -> Any:
