@@ -3,15 +3,15 @@ import sys
 import traceback
 from collections import OrderedDict, defaultdict
 
-from typing import Tuple, List, TypeVar, Set, Dict, Optional, TextIO
+from typing import Tuple, List, TypeVar, Set, Dict, Optional, TextIO, Callable
+from typing_extensions import Final
 
 from mypy.scope import Scope
 from mypy.options import Options
 from mypy.version import __version__ as mypy_version
-
-MYPY = False
-if MYPY:
-    from typing_extensions import Final
+from mypy.errorcodes import ErrorCode
+from mypy import errorcodes as codes
+from mypy.util import DEFAULT_SOURCE_OFFSET
 
 T = TypeVar('T')
 allowed_duplicates = ['@overload', 'Got:', 'Expected:']  # type: Final
@@ -48,6 +48,9 @@ class ErrorInfo:
     # The error message.
     message = ''
 
+    # The error code.
+    code = None  # type: Optional[ErrorCode]
+
     # If True, we should halt build after the file that generated this error.
     blocker = False
 
@@ -71,6 +74,7 @@ class ErrorInfo:
                  column: int,
                  severity: str,
                  message: str,
+                 code: Optional[ErrorCode],
                  blocker: bool,
                  only_once: bool,
                  origin: Optional[Tuple[str, int, int]] = None,
@@ -84,12 +88,23 @@ class ErrorInfo:
         self.column = column
         self.severity = severity
         self.message = message
+        self.code = code
         self.blocker = blocker
         self.only_once = only_once
         self.origin = origin or (file, line, line)
         self.target = target
 
 
+# Type used internally to represent errors:
+#   (path, line, column, severity, message, code)
+ErrorTuple = Tuple[Optional[str],
+                   int,
+                   int,
+                   str,
+                   str,
+                   Optional[ErrorCode]]
+
+
 class Errors:
     """Container for compile errors.
 
@@ -114,8 +129,9 @@ class Errors:
     # Path to current file.
     file = ''  # type: str
 
-    # Ignore errors on these lines of each file.
-    ignored_lines = None  # type: Dict[str, Set[int]]
+    # Ignore some errors on these lines of each file
+    # (path -> line -> error-codes)
+    ignored_lines = None  # type: Dict[str, Dict[int, List[str]]]
 
     # Lines on which an error was actually ignored.
     used_ignored_lines = None  # type: Dict[str, Set[int]]
@@ -138,17 +154,24 @@ class Errors:
     target_module = None  # type: Optional[str]
     scope = None  # type: Optional[Scope]
 
-    def __init__(self, show_error_context: bool = False,
-                 show_column_numbers: bool = False) -> None:
+    def __init__(self,
+                 show_error_context: bool = False,
+                 show_column_numbers: bool = False,
+                 show_error_codes: bool = False,
+                 pretty: bool = False,
+                 read_source: Optional[Callable[[str], Optional[List[str]]]] = None) -> None:
         self.show_error_context = show_error_context
         self.show_column_numbers = show_column_numbers
+        self.show_error_codes = show_error_codes
+        self.pretty = pretty
+        # We use fscache to read source code when showing snippets.
+        self.read_source = read_source
         self.initialize()
 
     def initialize(self) -> None:
         self.error_info_map = OrderedDict()
         self.flushed_files = set()
         self.import_ctx = []
-        self.type_name = [None]
         self.function_or_member = [None]
         self.ignored_lines = OrderedDict()
         self.used_ignored_lines = defaultdict(set)
@@ -161,10 +184,13 @@ class Errors:
         self.initialize()
 
     def copy(self) -> 'Errors':
-        new = Errors(self.show_error_context, self.show_column_numbers)
+        new = Errors(self.show_error_context,
+                     self.show_column_numbers,
+                     self.show_error_codes,
+                     self.pretty,
+                     self.read_source)
         new.file = self.file
         new.import_ctx = self.import_ctx[:]
-        new.type_name = self.type_name[:]
         new.function_or_member = self.function_or_member[:]
         new.target_module = self.target_module
         new.scope = self.scope
@@ -200,7 +226,7 @@ class Errors:
         self.scope = scope
 
     def set_file_ignored_lines(self, file: str,
-                               ignored_lines: Set[int],
+                               ignored_lines: Dict[int, List[str]],
                                ignore_all: bool = False) -> None:
         self.ignored_lines[file] = ignored_lines
         if ignore_all:
@@ -229,6 +255,8 @@ class Errors:
                line: int,
                column: Optional[int],
                message: str,
+               code: Optional[ErrorCode] = None,
+               *,
                blocker: bool = False,
                severity: str = 'error',
                file: Optional[str] = None,
@@ -240,7 +268,9 @@ class Errors:
 
         Args:
             line: line number of error
+            column: column number of error
             message: message to report
+            code: error code (defaults to 'misc'; not shown for notes)
             blocker: if True, don't continue analysis after this error
             severity: 'error' or 'note'
             file: if non-None, override current file as context
@@ -270,8 +300,10 @@ class Errors:
         if end_line is None:
             end_line = origin_line
 
+        code = code or codes.MISC
+
         info = ErrorInfo(self.import_context(), file, self.current_module(), type,
-                         function, line, column, severity, message,
+                         function, line, column, severity, message, code,
                          blocker, only_once,
                          origin=(self.file, origin_line, end_line),
                          target=self.current_target())
@@ -296,7 +328,7 @@ class Errors:
                 # Check each line in this context for "type: ignore" comments.
                 # line == end_line for most nodes, so we only loop once.
                 for scope_line in range(line, end_line + 1):
-                    if scope_line in self.ignored_lines[file]:
+                    if self.is_ignored_error(scope_line, info, self.ignored_lines[file]):
                         # Annotation requests us to ignore all errors on this line.
                         self.used_ignored_lines[file].add(scope_line)
                         return
@@ -308,6 +340,16 @@ class Errors:
             self.only_once_messages.add(info.message)
         self._add_error_info(file, info)
 
+    def is_ignored_error(self, line: int, info: ErrorInfo, ignores: Dict[int, List[str]]) -> bool:
+        if line not in ignores:
+            return False
+        elif not ignores[line]:
+            # Empty list means that we ignore all errors
+            return True
+        elif info.code:
+            return info.code.code in ignores[line]
+        return False
+
     def clear_errors_in_targets(self, path: str, targets: Set[str]) -> None:
         """Remove errors in specific fine-grained targets within a file."""
         if path in self.error_info_map:
@@ -322,11 +364,11 @@ class Errors:
     def generate_unused_ignore_errors(self, file: str) -> None:
         ignored_lines = self.ignored_lines[file]
         if not self.is_typeshed_file(file) and file not in self.ignored_files:
-            for line in ignored_lines - self.used_ignored_lines[file]:
+            for line in set(ignored_lines) - self.used_ignored_lines[file]:
                 # Don't use report since add_error_info will ignore the error!
                 info = ErrorInfo(self.import_context(), file, self.current_module(), None,
                                  None, line, -1, 'error', "unused 'type: ignore' comment",
-                                 False, False)
+                                 None, False, False)
                 self._add_error_info(file, info)
 
     def is_typeshed_file(self, file: str) -> bool:
@@ -368,28 +410,45 @@ class Errors:
                            use_stdout=True,
                            module_with_blocker=self.blocker_module())
 
-    def format_messages(self, error_info: List[ErrorInfo]) -> List[str]:
+    def format_messages(self, error_info: List[ErrorInfo],
+                        source_lines: Optional[List[str]]) -> List[str]:
         """Return a string list that represents the error messages.
 
-        Use a form suitable for displaying to the user.
+        Use a form suitable for displaying to the user. If self.pretty
+        is True also append a relevant trimmed source code line (only for
+        severity 'error').
         """
         a = []  # type: List[str]
         errors = self.render_messages(self.sort_messages(error_info))
         errors = self.remove_duplicates(errors)
-        for file, line, column, severity, message in errors:
+        for file, line, column, severity, message, code in errors:
             s = ''
             if file is not None:
-                if self.show_column_numbers and line is not None and line >= 0 \
-                        and column is not None and column >= 0:
+                if self.show_column_numbers and line >= 0 and column >= 0:
                     srcloc = '{}:{}:{}'.format(file, line, 1 + column)
-                elif line is not None and line >= 0:
+                elif line >= 0:
                     srcloc = '{}:{}'.format(file, line)
                 else:
                     srcloc = file
                 s = '{}: {}: {}'.format(srcloc, severity, message)
             else:
                 s = message
+            if self.show_error_codes and code and severity != 'note':
+                # If note has an error code, it is related to a previous error. Avoid
+                # displaying duplicate error codes.
+                s = '{}  [{}]'.format(s, code.code)
             a.append(s)
+            if self.pretty:
+                # Add source code fragment and a location marker.
+                if severity == 'error' and source_lines and line > 0:
+                    source_line = source_lines[line - 1]
+                    if column < 0:
+                        # Something went wrong, take first non-empty column.
+                        column = len(source_line) - len(source_line.lstrip())
+                    # Note, currently coloring uses the offset to detect source snippets,
+                    # so these offsets should not be arbitrary.
+                    a.append(' ' * DEFAULT_SOURCE_OFFSET + source_line)
+                    a.append(' ' * (DEFAULT_SOURCE_OFFSET + column) + '^')
         return a
 
     def file_messages(self, path: str) -> List[str]:
@@ -400,7 +459,11 @@ class Errors:
         if path not in self.error_info_map:
             return []
         self.flushed_files.add(path)
-        return self.format_messages(self.error_info_map[path])
+        source_lines = None
+        if self.pretty:
+            assert self.read_source
+            source_lines = self.read_source(path)
+        return self.format_messages(self.error_info_map[path], source_lines)
 
     def new_messages(self) -> List[str]:
         """Return a string list of new error messages.
@@ -424,18 +487,16 @@ class Errors:
                    for info in errs
                    if info.target)
 
-    def render_messages(self, errors: List[ErrorInfo]) -> List[Tuple[Optional[str], int, int,
-                                                                     str, str]]:
+    def render_messages(self,
+                        errors: List[ErrorInfo]) -> List[ErrorTuple]:
         """Translate the messages into a sequence of tuples.
 
-        Each tuple is of form (path, line, col, severity, message).
+        Each tuple is of form (path, line, col, severity, message, code).
         The rendered sequence includes information about error contexts.
         The path item may be None. If the line item is negative, the
         line number is not defined for the tuple.
         """
-        result = []  # type: List[Tuple[Optional[str], int, int, str, str]]
-        # (path, line, column, severity, message)
-
+        result = []  # type: List[ErrorTuple]
         prev_import_context = []  # type: List[Tuple[str, int]]
         prev_function_or_member = None  # type: Optional[str]
         prev_type = None  # type: Optional[str]
@@ -459,7 +520,7 @@ class Errors:
                     # Remove prefix to ignore from path (if present) to
                     # simplify path.
                     path = remove_path_prefix(path, self.ignore_prefix)
-                    result.append((None, -1, -1, 'note', fmt.format(path, line)))
+                    result.append((None, -1, -1, 'note', fmt.format(path, line), None))
                     i -= 1
 
             file = self.simplify_path(e.file)
@@ -471,27 +532,27 @@ class Errors:
                     e.type != prev_type):
                 if e.function_or_member is None:
                     if e.type is None:
-                        result.append((file, -1, -1, 'note', 'At top level:'))
+                        result.append((file, -1, -1, 'note', 'At top level:', None))
                     else:
                         result.append((file, -1, -1, 'note', 'In class "{}":'.format(
-                            e.type)))
+                            e.type), None))
                 else:
                     if e.type is None:
                         result.append((file, -1, -1, 'note',
                                        'In function "{}":'.format(
-                                           e.function_or_member)))
+                                           e.function_or_member), None))
                     else:
                         result.append((file, -1, -1, 'note',
                                        'In member "{}" of class "{}":'.format(
-                                           e.function_or_member, e.type)))
+                                           e.function_or_member, e.type), None))
             elif e.type != prev_type:
                 if e.type is None:
-                    result.append((file, -1, -1, 'note', 'At top level:'))
+                    result.append((file, -1, -1, 'note', 'At top level:', None))
                 else:
                     result.append((file, -1, -1, 'note',
-                                   'In class "{}":'.format(e.type)))
+                                   'In class "{}":'.format(e.type), None))
 
-            result.append((file, e.line, e.column, e.severity, e.message))
+            result.append((file, e.line, e.column, e.severity, e.message, e.code))
 
             prev_import_context = e.import_ctx
             prev_function_or_member = e.function_or_member
@@ -522,10 +583,9 @@ class Errors:
             result.extend(a)
         return result
 
-    def remove_duplicates(self, errors: List[Tuple[Optional[str], int, int, str, str]]
-                          ) -> List[Tuple[Optional[str], int, int, str, str]]:
+    def remove_duplicates(self, errors: List[ErrorTuple]) -> List[ErrorTuple]:
         """Remove duplicates from a sorted error list."""
-        res = []  # type: List[Tuple[Optional[str], int, int, str, str]]
+        res = []  # type: List[ErrorTuple]
         i = 0
         while i < len(errors):
             dup = False
@@ -619,7 +679,7 @@ def report_internal_error(err: Exception,
     # Print "INTERNAL ERROR" message.
     print('{}error: INTERNAL ERROR --'.format(prefix),
           'Please try using mypy master on Github:\n'
-          'https://mypy.rtfd.io/en/latest/common_issues.html#using-development-mypy-build',
+          'https://mypy.rtfd.io/en/latest/common_issues.html#using-a-development-mypy-build',
           file=stderr)
     if options.show_traceback:
         print('Please report a bug at https://github.com/python/mypy/issues',
