@@ -76,9 +76,10 @@ class NodeFixer(NodeVisitor[None]):
                     stnode = lookup_qualified_stnode(self.modules, cross_ref,
                                                      self.allow_missing)
                     if stnode is not None:
+                        assert stnode.node is not None
                         value.node = stnode.node
                     elif not self.allow_missing:
-                        assert stnode is not None, "Could not find cross-ref %s" % (cross_ref,)
+                        assert False, "Could not find cross-ref %s" % (cross_ref,)
                     else:
                         # We have a missing crossref in allow missing mode, need to put something
                         value.node = missing_info(self.modules)
@@ -185,6 +186,10 @@ class TypeFixer(TypeVisitor[None]):
         for ct in t.items():
             ct.accept(self)
 
+    def visit_erased_type(self, o: Any) -> None:
+        # This type should exist only temporarily during type inference
+        raise RuntimeError("Shouldn't get here", o)
+
     def visit_deleted_type(self, o: Any) -> None:
         pass  # Nothing to descend into.
 
@@ -209,6 +214,12 @@ class TypeFixer(TypeVisitor[None]):
             for it in tdt.items.values():
                 it.accept(self)
         if tdt.fallback is not None:
+            if tdt.fallback.type_ref is not None:
+                if lookup_qualified(self.modules, tdt.fallback.type_ref,
+                                    self.allow_missing) is None:
+                    # We reject fake TypeInfos for TypedDict fallbacks because
+                    # the latter are used in type checking and must be valid.
+                    tdt.fallback.type_ref = 'typing._TypedDict'
             tdt.fallback.accept(self)
 
     def visit_literal_type(self, lt: LiteralType) -> None:
