@@ -1,13 +1,14 @@
 """Hack for handling non-mypyc compiled plugins with a mypyc-compiled mypy"""
 
 from typing import Optional, Callable, Any, Dict, List, Tuple
+from typing_extensions import TYPE_CHECKING
+
 from mypy.options import Options
 from mypy.types import Type, CallableType
 from mypy.nodes import SymbolTableNode, MypyFile
 from mypy.lookup import lookup_fully_qualified
 
-MYPY = False
-if MYPY:
+if TYPE_CHECKING:
     import mypy.plugin
 
 
@@ -21,7 +22,8 @@ class InterpretedPlugin:
     that proxies to this interpreted version.
     """
 
-    def __new__(cls, *args: Any, **kwargs: Any) -> 'mypy.plugin.Plugin':
+    # ... mypy doesn't like these shenanigans so we have to type ignore it!
+    def __new__(cls, *args: Any, **kwargs: Any) -> 'mypy.plugin.Plugin':  # type: ignore
         from mypy.plugin import WrapperPlugin
         plugin = object.__new__(cls)
         plugin.__init__(*args, **kwargs)
