@@ -1,22 +1,20 @@
 """Plugin that provides support for dataclasses."""
 
 from collections import OrderedDict
-from typing import Dict, List, Set, Tuple
+
+from typing import Dict, List, Set, Tuple, Optional
+from typing_extensions import Final
 
 from mypy.nodes import (
     ARG_OPT, ARG_POS, MDEF, Argument, AssignmentStmt, CallExpr,
     Context, Expression, FuncDef, JsonDict, NameExpr, RefExpr,
-    SymbolTableNode, TempNode, TypeInfo, Var, TypeVarExpr
+    SymbolTableNode, TempNode, TypeInfo, Var, TypeVarExpr, PlaceholderNode
 )
 from mypy.plugin import ClassDefContext
 from mypy.plugins.common import add_method, _get_decorator_bool_argument
-from mypy.types import Instance, NoneType, TypeVarDef, TypeVarType
+from mypy.types import Instance, NoneType, TypeVarDef, TypeVarType, get_proper_type
 from mypy.server.trigger import make_wildcard_trigger
 
-MYPY = False
-if MYPY:
-    from typing_extensions import Final
-
 # The set of decorators that generate dataclasses.
 dataclass_makers = {
     'dataclass',
@@ -81,12 +79,18 @@ class DataclassTransformer:
         ctx = self._ctx
         info = self._ctx.cls.info
         attributes = self.collect_attributes()
-        if ctx.api.options.new_semantic_analyzer:
-            # Check if attribute types are ready.
-            for attr in attributes:
-                if info[attr.name].type is None:
-                    ctx.api.defer()
-                    return
+        if attributes is None:
+            # Some definitions are not ready, defer() should be already called.
+            return
+        for attr in attributes:
+            node = info.get(attr.name)
+            if node is None:
+                # Nodes of superclass InitVars not used in __init__ cannot be reached.
+                assert attr.is_init_var and not attr.is_in_init
+                continue
+            if node.type is None:
+                ctx.api.defer()
+                return
         decorator_arguments = {
             'init': _get_decorator_bool_argument(self._ctx, 'init', True),
             'eq': _get_decorator_bool_argument(self._ctx, 'eq', True),
@@ -94,7 +98,13 @@ class DataclassTransformer:
             'frozen': _get_decorator_bool_argument(self._ctx, 'frozen', False),
         }
 
-        if decorator_arguments['init']:
+        # If there are no attributes, it may be that the new semantic analyzer has not
+        # processed them yet. In order to work around this, we can simply skip generating
+        # __init__ if there are no attributes, because if the user truly did not define any,
+        # then the object default __init__ with an empty signature will be present anyway.
+        if (decorator_arguments['init'] and
+                ('__init__' not in info.names or info.names['__init__'].plugin_generated) and
+                attributes):
             add_method(
                 ctx,
                 '__init__',
@@ -179,7 +189,11 @@ class DataclassTransformer:
         """Remove init-only vars from the class and reset init var declarations."""
         for attr in attributes:
             if attr.is_init_var:
-                del info.names[attr.name]
+                if attr.name in info.names:
+                    del info.names[attr.name]
+                else:
+                    # Nodes of superclass InitVars not used in __init__ cannot be reached.
+                    assert attr.is_init_var and not attr.is_in_init
                 for stmt in info.defn.defs.body:
                     if isinstance(stmt, AssignmentStmt) and stmt.unanalyzed_type:
                         lvalue = stmt.lvalues[0]
@@ -188,7 +202,7 @@ class DataclassTransformer:
                             # recreate a symbol node for this attribute.
                             lvalue.node = None
 
-    def collect_attributes(self) -> List[DataclassAttribute]:
+    def collect_attributes(self) -> Optional[List[DataclassAttribute]]:
         """Collect all attributes declared in the dataclass and its parents.
 
         All assignments of the form
@@ -215,7 +229,16 @@ class DataclassTransformer:
             if not isinstance(lhs, NameExpr):
                 continue
 
-            node = cls.info.names[lhs.name].node
+            sym = cls.info.names.get(lhs.name)
+            if sym is None:
+                # This name is likely blocked by a star import. We don't need to defer because
+                # defer() is already called by mark_incomplete().
+                continue
+
+            node = sym.node
+            if isinstance(node, PlaceholderNode):
+                # This node is not ready yet.
+                return None
             assert isinstance(node, Var)
 
             # x: ClassVar[int] is ignored by dataclasses.
@@ -224,12 +247,11 @@ class DataclassTransformer:
 
             # x: InitVar[int] is turned into x: int and is removed from the class.
             is_init_var = False
-            if (
-                    isinstance(node.type, Instance) and
-                    node.type.type.fullname() == 'dataclasses.InitVar'
-            ):
+            node_type = get_proper_type(node.type)
+            if (isinstance(node_type, Instance) and
+                    node_type.type.fullname() == 'dataclasses.InitVar'):
                 is_init_var = True
-                node.type = node.type.args[0]
+                node.type = node_type.args[0]
 
             has_field_call, field_args = _collect_field_args(stmt.rvalue)
 
