@@ -114,7 +114,7 @@ test cases (test-data/unit/fine-grained*.test).
 
 import time
 from typing import (
-    Dict, List, Set, Tuple, Union, Optional, NamedTuple, Callable, Sequence
+    Dict, List, Set, Tuple, Union, Optional, NamedTuple, Sequence
 )
 from typing_extensions import Final
 
@@ -132,15 +132,13 @@ from mypy.nodes import (
 )
 from mypy.options import Options
 from mypy.fscache import FileSystemCache
-from mypy.semanal import apply_semantic_analyzer_patches
 from mypy.server.astdiff import (
     snapshot_symbol_table, compare_symbol_table_snapshots, SnapshotItem
 )
-from mypy.newsemanal.semanal_main import semantic_analysis_for_scc, semantic_analysis_for_targets
+from mypy.semanal_main import semantic_analysis_for_scc, semantic_analysis_for_targets
 from mypy.server.astmerge import merge_asts
-from mypy.server.aststrip import strip_target
-from mypy.server.aststripnew import strip_target_new, SavedAttributes
-from mypy.server.deps import get_dependencies_of_target
+from mypy.server.aststrip import strip_target, SavedAttributes
+from mypy.server.deps import get_dependencies_of_target, merge_dependencies
 from mypy.server.target import trigger_to_target
 from mypy.server.trigger import make_trigger, WILDCARD_TAG
 from mypy.util import module_prefix, split_target
@@ -163,7 +161,9 @@ class FineGrainedBuildManager:
         self.manager = manager
         self.graph = result.graph
         self.previous_modules = get_module_to_path_map(self.graph)
-        self.deps = get_all_dependencies(manager, self.graph)
+        self.deps = manager.fg_deps
+        # Merge in any root dependencies that may not have been loaded
+        merge_dependencies(manager.load_fine_grained_deps(FAKE_ROOT_MODULE), self.deps)
         self.previous_targets_with_errors = manager.errors.targets()
         self.previous_messages = result.errors[:]
         # Module, if any, that had blocking errors in the last run as (id, path) tuple.
@@ -378,7 +378,8 @@ class FineGrainedBuildManager:
             self.manager.log_fine_grained('triggered: %r' % sorted(filtered))
         self.triggered.extend(triggered | self.previous_targets_with_errors)
         if module in graph:
-            merge_dependencies(graph[module].compute_fine_grained_deps(), self.deps)
+            graph[module].update_fine_grained_deps(self.deps)
+            graph[module].free_state()
         remaining += propagate_changes_using_dependencies(
             manager, graph, self.deps, triggered,
             {module},
@@ -453,15 +454,11 @@ def ensure_trees_loaded(manager: BuildManager, graph: Dict[str, State],
         process_fresh_modules(graph, to_process, manager)
 
 
-def get_all_dependencies(manager: BuildManager, graph: Dict[str, State]) -> Dict[str, Set[str]]:
-    """Return the fine-grained dependency map for an entire build."""
-    # Deps for each module were computed during build() or loaded from the cache.
-    deps = manager.load_fine_grained_deps(FAKE_ROOT_MODULE)  # type: Dict[str, Set[str]]
-    for id in graph:
-        if graph[id].tree is not None:
-            merge_dependencies(graph[id].compute_fine_grained_deps(), deps)
-    TypeState.add_all_protocol_deps(deps)
-    return deps
+def fix_fg_dependencies(manager: BuildManager, deps: Dict[str, Set[str]]) -> None:
+    """Populate the dependencies with stuff that build may have missed"""
+    # This means the root module and typestate
+    merge_dependencies(manager.load_fine_grained_deps(FAKE_ROOT_MODULE), deps)
+    # TypeState.add_all_protocol_deps(deps)
 
 
 # The result of update_module_isolated when no blockers, with these items:
@@ -578,25 +575,12 @@ def update_module_isolated(module: str,
     assert state.tree is not None, "file must be at least parsed"
     t0 = time.time()
     # TODO: state.fix_suppressed_dependencies()?
-    if not manager.options.new_semantic_analyzer:
-        if module == 'typing':
-            # We need to manually add typing aliases to builtins, like we
-            # do in process_stale_scc. Because this can't be done until
-            # builtins is also loaded, there isn't an obvious way to
-            # refactor this.
-            manager.semantic_analyzer.add_builtin_aliases(state.tree)
     try:
-        if not manager.options.new_semantic_analyzer:
-            state.semantic_analysis()
-        else:
-            semantic_analysis_for_scc(graph, [state.id], manager.errors)
+        semantic_analysis_for_scc(graph, [state.id], manager.errors)
     except CompileError as err:
         # There was a blocking error, so module AST is incomplete. Restore old modules.
         restore([module])
         return BlockedUpdate(module, path, remaining_modules, err.messages)
-    if not manager.options.new_semantic_analyzer:
-        state.semantic_analysis_pass_three()
-        state.semantic_analysis_apply_patches()
 
     # Merge old and new ASTs.
     new_modules_dict = {module: state.tree}  # type: Dict[str, Optional[MypyFile]]
@@ -608,15 +592,12 @@ def update_module_isolated(module: str,
     state.type_check_first_pass()
     state.type_check_second_pass()
     t2 = time.time()
-    state.compute_fine_grained_deps()
-    t3 = time.time()
     state.finish_passes()
-    t4 = time.time()
+    t3 = time.time()
     manager.add_stats(
         semanal_time=t1 - t0,
         typecheck_time=t2 - t1,
-        deps_time=t3 - t2,
-        finish_passes_time=t4 - t3)
+        finish_passes_time=t3 - t2)
 
     graph[module] = state
 
@@ -700,14 +681,6 @@ def get_sources(fscache: FileSystemCache,
     return sources
 
 
-def merge_dependencies(new_deps: Dict[str, Set[str]],
-                       deps: Dict[str, Set[str]]) -> None:
-    for trigger, targets in new_deps.items():
-        deps.setdefault(trigger, set()).update(targets)
-    # Merge also the newly added protocol deps.
-    TypeState.update_protocol_deps(deps)
-
-
 def calculate_active_triggers(manager: BuildManager,
                               old_snapshots: Dict[str, Dict[str, SnapshotItem]],
                               new_modules: Dict[str, Optional[MypyFile]]) -> Set[str]:
@@ -950,14 +923,8 @@ def reprocess_nodes(manager: BuildManager,
     saved_attrs = {}  # type: SavedAttributes
     for deferred in nodes:
         processed_targets.append(deferred.node.fullname())
-        if not manager.options.new_semantic_analyzer:
-            strip_target(deferred.node)
-        else:
-            strip_target_new(deferred.node, saved_attrs)
-    if not options.new_semantic_analyzer:
-        re_analyze_nodes(file_node, nodes, manager, options)
-    else:
-        semantic_analysis_for_targets(graph[module_id], nodes, graph, saved_attrs)
+        strip_target(deferred.node, saved_attrs)
+    semantic_analysis_for_targets(graph[module_id], nodes, graph, saved_attrs)
     # Merge symbol tables to preserve identities of AST nodes. The file node will remain
     # the same, but other nodes may have been recreated with different identities, such as
     # NamedTuples defined using assignment statements.
@@ -994,46 +961,11 @@ def reprocess_nodes(manager: BuildManager,
     # Report missing imports.
     graph[module_id].verify_dependencies()
 
+    graph[module_id].free_state()
+
     return new_triggered
 
 
-def re_analyze_nodes(file_node: MypyFile, nodes: List[FineGrainedDeferredNode],
-                     manager: BuildManager, options: Options) -> None:
-    """Perform semantic analysis of targets in a given file.
-
-    This uses the old semantic analyzer.
-    """
-    semantic_analyzer = manager.semantic_analyzer
-    patches = []  # type: List[Tuple[int, Callable[[], None]]]
-
-    # Second pass of semantic analysis. We don't redo the first pass, because it only
-    # does local things that won't go stale.
-    for deferred in nodes:
-        with semantic_analyzer.file_context(
-                file_node=file_node,
-                fnam=file_node.path,
-                options=options,
-                active_type=deferred.active_typeinfo):
-            manager.semantic_analyzer.refresh_partial(deferred.node, patches)
-
-    # Third pass of semantic analysis.
-    for deferred in nodes:
-        with semantic_analyzer.file_context(
-                file_node=file_node,
-                fnam=file_node.path,
-                options=options,
-                active_type=deferred.active_typeinfo,
-                scope=manager.semantic_analyzer_pass3.scope):
-            manager.semantic_analyzer_pass3.refresh_partial(deferred.node, patches)
-
-    with semantic_analyzer.file_context(
-            file_node=file_node,
-            fnam=file_node.path,
-            options=options,
-            active_type=None):
-        apply_semantic_analyzer_patches(patches)
-
-
 def find_symbol_tables_recursive(prefix: str, symbols: SymbolTable) -> Dict[str, SymbolTable]:
     """Find all nested symbol tables.
 
@@ -1096,11 +1028,9 @@ def lookup_target(manager: BuildManager,
     node = modules[module]  # type: Optional[SymbolNode]
     file = None  # type: Optional[MypyFile]
     active_class = None
-    active_class_name = None
     for c in components:
         if isinstance(node, TypeInfo):
             active_class = node
-            active_class_name = node.name()
         if isinstance(node, MypyFile):
             file = node
         if (not isinstance(node, (MypyFile, TypeInfo))
@@ -1125,7 +1055,7 @@ def lookup_target(manager: BuildManager,
             # a deserialized TypeInfo with missing attributes.
             not_found()
             return [], None
-        result = [FineGrainedDeferredNode(file, None, None)]
+        result = [FineGrainedDeferredNode(file, None)]
         stale_info = None  # type: Optional[TypeInfo]
         if node.is_protocol:
             stale_info = node
@@ -1150,7 +1080,7 @@ def lookup_target(manager: BuildManager,
         # context will be wrong and it could be a partially initialized deserialized node.
         not_found()
         return [], None
-    return [FineGrainedDeferredNode(node, active_class_name, active_class)], None
+    return [FineGrainedDeferredNode(node, active_class)], None
 
 
 def is_verbose(manager: BuildManager) -> bool:
