@@ -92,7 +92,7 @@ class StrConv(NodeVisitor[str]):
         # Omit path to special file with name "main". This is used to simplify
         # test case descriptions; the file "main" is used by default in many
         # test cases.
-        if o.path is not None and o.path != 'main':
+        if o.path != 'main':
             # Insert path. Normalize directory separators to / to unify test
             # case# output in all platforms.
             a.insert(0, o.path.replace(os.sep, '/'))
@@ -425,6 +425,9 @@ class StrConv(NodeVisitor[str]):
             # REVEAL_LOCALS
             return self.dump([o.local_nodes], o)
 
+    def visit_assignment_expr(self, o: 'mypy.nodes.AssignmentExpr') -> str:
+        return self.dump([o.target, o.value], o)
+
     def visit_unary_expr(self, o: 'mypy.nodes.UnaryExpr') -> str:
         return self.dump([o.op, o.expr], o)
 
