@@ -25,16 +25,20 @@ Other things:
 from typing import (
     List, Optional, Tuple, Dict, Callable, Union, NamedTuple, TypeVar, Iterator,
 )
+from typing_extensions import TypedDict
 
 from mypy.state import strict_optional_set
 from mypy.types import (
-    Type, AnyType, TypeOfAny, CallableType, UnionType, NoneType, Instance, TupleType, is_optional,
-    TypeStrVisitor,
+    Type, AnyType, TypeOfAny, CallableType, UnionType, NoneType, Instance, TupleType,
+    TypeVarType, FunctionLike,
+    TypeStrVisitor, TypeTranslator,
+    is_optional, remove_optional, ProperType, get_proper_type,
 )
 from mypy.build import State, Graph
 from mypy.nodes import (
     ARG_STAR, ARG_NAMED, ARG_STAR2, ARG_NAMED_OPT, FuncDef, MypyFile, SymbolTable,
-    SymbolNode, TypeInfo, Expression, ReturnStmt,
+    Decorator, RefExpr,
+    SymbolNode, TypeInfo, Expression, ReturnStmt, CallExpr,
     reverse_builtin_aliases,
 )
 from mypy.server.update import FineGrainedBuildManager
@@ -45,6 +49,7 @@ from mypy.checkexpr import has_any_type
 
 from mypy.join import join_type_list
 from mypy.sametypes import is_same_type
+from mypy.typeops import make_simplified_union
 
 from contextlib import contextmanager
 
@@ -52,6 +57,11 @@ import itertools
 import json
 import os
 
+
+PyAnnotateSignature = TypedDict('PyAnnotateSignature',
+                                {'return_type': str, 'arg_types': List[str]})
+
+
 Callsite = NamedTuple(
     'Callsite',
     [('path', str),
@@ -134,6 +144,7 @@ def is_explicit_any(typ: AnyType) -> bool:
 
 
 def is_implicit_any(typ: Type) -> bool:
+    typ = get_proper_type(typ)
     return isinstance(typ, AnyType) and not is_explicit_any(typ)
 
 
@@ -141,9 +152,12 @@ class SuggestionEngine:
     """Engine for finding call sites and suggesting signatures."""
 
     def __init__(self, fgmanager: FineGrainedBuildManager,
+                 *,
                  json: bool,
                  no_errors: bool = False,
-                 no_any: bool = False) -> None:
+                 no_any: bool = False,
+                 try_text: bool = False,
+                 flex_any: Optional[float] = None) -> None:
         self.fgmanager = fgmanager
         self.manager = fgmanager.manager
         self.plugin = self.manager.plugin
@@ -151,7 +165,10 @@ class SuggestionEngine:
 
         self.give_json = json
         self.no_errors = no_errors
-        self.no_any = no_any
+        self.try_text = try_text
+        self.flex_any = flex_any
+        if no_any:
+            self.flex_any = 1.0
 
         self.max_guesses = 16
 
@@ -164,7 +181,7 @@ class SuggestionEngine:
         if self.give_json:
             return self.json_suggestion(function, suggestion)
         else:
-            return suggestion
+            return self.format_signature(suggestion)
 
     def suggest_callsites(self, function: str) -> str:
         """Find a list of call sites of function."""
@@ -220,8 +237,9 @@ class SuggestionEngine:
         types = []  # type: List[List[Type]]
         for i in range(len(base.arg_kinds)):
             # Make self args Any but this will get overriden somewhere in the checker
+            # We call this a special form so that has_any_type doesn't consider it to be a real any
             if i == 0 and is_method:
-                types.append([AnyType(TypeOfAny.explicit)])
+                types.append([AnyType(TypeOfAny.special_form)])
                 continue
 
             all_arg_types = []
@@ -235,7 +253,11 @@ class SuggestionEngine:
             if default:
                 all_arg_types.append(default)
 
-            if all_arg_types:
+            if len(all_arg_types) == 1 and isinstance(get_proper_type(all_arg_types[0]), NoneType):
+                types.append(
+                    [UnionType.make_union([all_arg_types[0],
+                                           AnyType(TypeOfAny.explicit)])])
+            elif all_arg_types:
                 types.append(generate_type_combinations(all_arg_types))
             else:
                 # If we don't have anything, we'll try Any and object
@@ -248,6 +270,12 @@ class SuggestionEngine:
         return [self.manager.all_types[arg.initializer] if arg.initializer else None
                 for arg in fdef.arguments]
 
+    def add_adjustments(self, typs: List[Type]) -> List[Type]:
+        if not self.try_text or self.manager.options.python_version[0] != 2:
+            return typs
+        translator = StrToText(self.builtin_type)
+        return dedup(typs + [tp.accept(translator) for tp in typs])
+
     def get_guesses(self, is_method: bool, base: CallableType, defaults: List[Optional[Type]],
                     callsites: List[Callsite]) -> List[CallableType]:
         """Compute a list of guesses for a function's type.
@@ -255,6 +283,7 @@ class SuggestionEngine:
         This focuses just on the argument types, and doesn't change the provided return type.
         """
         options = self.get_args(is_method, base, defaults, callsites)
+        options = [self.add_adjustments(tps) for tps in options]
         return [base.copy_modified(arg_types=list(x)) for x in itertools.product(*options)]
 
     def get_callsites(self, func: FuncDef) -> Tuple[List[Callsite], List[str]]:
@@ -271,13 +300,13 @@ class SuggestionEngine:
 
         return collector_plugin.mystery_hits, errors
 
-    def filter_options(self, guesses: List[CallableType]) -> List[CallableType]:
+    def filter_options(self, guesses: List[CallableType], is_method: bool) -> List[CallableType]:
         """Apply any configured filters to the possible guesses.
 
-        Currently the only option is disabling Anys."""
+        Currently the only option is filtering based on Any prevalance."""
         return [
             t for t in guesses
-            if not self.no_any or not callable_has_any(t)
+            if self.flex_any is None or any_score_callable(t, is_method) >= self.flex_any
         ]
 
     def find_best(self, func: FuncDef, guesses: List[CallableType]) -> Tuple[CallableType, int]:
@@ -289,10 +318,10 @@ class SuggestionEngine:
             raise SuggestionFailure("No guesses that match criteria!")
         errors = {guess: self.try_type(func, guess) for guess in guesses}
         best = min(guesses,
-                   key=lambda s: (count_errors(errors[s]), score_callable(s)))
+                   key=lambda s: (count_errors(errors[s]), self.score_callable(s)))
         return best, count_errors(errors[best])
 
-    def get_suggestion(self, function: str) -> str:
+    def get_suggestion(self, function: str) -> PyAnnotateSignature:
         """Compute a suggestion for a function.
 
         Return the type and whether the first argument should be ignored.
@@ -304,8 +333,10 @@ class SuggestionEngine:
         if self.no_errors and orig_errors:
             raise SuggestionFailure("Function does not typecheck.")
 
-        # FIXME: what about static and class methods?
-        is_method = bool(node.info)
+        is_method = bool(node.info) and not node.is_static
+
+        if len(node.arg_names) >= 10:
+            raise SuggestionFailure("Too many arguments")
 
         with strict_optional_set(graph[mod].options.strict_optional):
             guesses = self.get_guesses(
@@ -313,7 +344,7 @@ class SuggestionEngine:
                 self.get_trivial_type(node),
                 self.get_default_arg_types(graph[mod], node),
                 callsites)
-        guesses = self.filter_options(guesses)
+        guesses = self.filter_options(guesses, is_method)
         if len(guesses) > self.max_guesses:
             raise SuggestionFailure("Too many possibilities!")
         best, _ = self.find_best(node, guesses)
@@ -328,13 +359,13 @@ class SuggestionEngine:
                 ret_types = [NoneType()]
 
         guesses = [best.copy_modified(ret_type=t) for t in ret_types]
-        guesses = self.filter_options(guesses)
+        guesses = self.filter_options(guesses, is_method)
         best, errors = self.find_best(node, guesses)
 
         if self.no_errors and errors:
             raise SuggestionFailure("No annotation without errors")
 
-        return self.format_callable(mod, is_method, best)
+        return self.pyannotate_signature(mod, is_method, best)
 
     def format_args(self,
                     arg_kinds: List[List[int]],
@@ -388,12 +419,39 @@ class SuggestionEngine:
             raise SuggestionFailure("Unknown %s %s" %
                                     ("method" if len(components) > 1 else "function", key))
         node = names[funcname].node
+        if isinstance(node, Decorator):
+            node = self.extract_from_decorator(node)
+            if not node:
+                raise SuggestionFailure("Object %s is a decorator we can't handle" % key)
+
         if not isinstance(node, FuncDef):
             raise SuggestionFailure("Object %s is not a function" % key)
 
         return (modname, tail, node)
 
-    def try_type(self, func: FuncDef, typ: Type) -> List[str]:
+    def extract_from_decorator(self, node: Decorator) -> Optional[FuncDef]:
+        for dec in node.decorators:
+            typ = None
+            if (isinstance(dec, RefExpr)
+                    and isinstance(dec.node, FuncDef)):
+                typ = dec.node.type
+            elif (isinstance(dec, CallExpr)
+                    and isinstance(dec.callee, RefExpr)
+                    and isinstance(dec.callee.node, FuncDef)
+                    and isinstance(dec.callee.node.type, CallableType)):
+                typ = get_proper_type(dec.callee.node.type.ret_type)
+
+            if not isinstance(typ, FunctionLike):
+                return None
+            for ct in typ.items():
+                if not (len(ct.arg_types) == 1
+                        and isinstance(ct.arg_types[0], TypeVarType)
+                        and ct.arg_types[0] == ct.ret_type):
+                    return None
+
+        return node.func
+
+    def try_type(self, func: FuncDef, typ: ProperType) -> List[str]:
         """Recheck a function while assuming it has type typ.
 
         Return all error messages.
@@ -433,14 +491,14 @@ class SuggestionEngine:
         return state.tree
 
     def builtin_type(self, s: str) -> Instance:
-        if self.manager.options.new_semantic_analyzer:
-            return self.manager.new_semantic_analyzer.builtin_type(s)
-        else:
-            return self.manager.semantic_analyzer.builtin_type(s)
+        return self.manager.semantic_analyzer.builtin_type(s)
 
-    def json_suggestion(self, function: str, suggestion: str) -> str:
+    def json_suggestion(self, function: str, suggestion: PyAnnotateSignature) -> str:
         """Produce a json blob for a suggestion suitable for application by pyannotate."""
         mod, func_name, node = self.find_node(function)
+        # pyannotate irritatingly drops class names for class and static methods
+        if node.is_class or node.is_static:
+            func_name = func_name.split('.', 1)[-1]
 
         # pyannotate works with either paths relative to where the
         # module is rooted or with absolute paths. We produce absolute
@@ -448,7 +506,7 @@ class SuggestionEngine:
         path = os.path.abspath(self.graph[mod].xpath)
 
         obj = {
-            'type_comments': [suggestion],
+            'signature': suggestion,
             'line': node.line,
             'path': path,
             'func_name': func_name,
@@ -456,34 +514,108 @@ class SuggestionEngine:
         }
         return json.dumps([obj], sort_keys=True)
 
-    def format_callable(self,
-                        cur_module: Optional[str], is_method: bool, typ: CallableType) -> str:
-        """Format a callable type in a way suitable as an annotation... kind of"""
+    def pyannotate_signature(
+        self,
+        cur_module: Optional[str],
+        is_method: bool,
+        typ: CallableType
+    ) -> PyAnnotateSignature:
+        """Format a callable type as a pyannotate dict"""
         start = int(is_method)
-        s = "({}) -> {}".format(
-            ", ".join([self.format_type(cur_module, t) for t in typ.arg_types[start:]]),
-            self.format_type(cur_module, typ.ret_type))
-        return s
+        return {
+            'arg_types': [self.format_type(cur_module, t) for t in typ.arg_types[start:]],
+            'return_type': self.format_type(cur_module, typ.ret_type),
+        }
+
+    def format_signature(self, sig: PyAnnotateSignature) -> str:
+        """Format a callable type in a way suitable as an annotation... kind of"""
+        return "({}) -> {}".format(
+            ", ".join(sig['arg_types']),
+            sig['return_type']
+        )
 
     def format_type(self, cur_module: Optional[str], typ: Type) -> str:
         return typ.accept(TypeFormatter(cur_module, self.graph))
 
+    def score_type(self, t: Type, arg_pos: bool) -> int:
+        """Generate a score for a type that we use to pick which type to use.
+
+        Lower is better, prefer non-union/non-any types. Don't penalize optionals.
+        """
+        t = get_proper_type(t)
+        if isinstance(t, AnyType):
+            return 20
+        if arg_pos and isinstance(t, NoneType):
+            return 20
+        if isinstance(t, UnionType):
+            if any(isinstance(x, AnyType) for x in t.items):
+                return 20
+            if not is_optional(t):
+                return 10
+        if isinstance(t, CallableType) and (has_any_type(t) or is_tricky_callable(t)):
+            return 10
+        if self.try_text and isinstance(t, Instance) and t.type.fullname() == 'builtins.str':
+            return 1
+        return 0
+
+    def score_callable(self, t: CallableType) -> int:
+        return (sum([self.score_type(x, arg_pos=True) for x in t.arg_types]) +
+                self.score_type(t.ret_type, arg_pos=False))
+
+
+def any_score_type(ut: Type, arg_pos: bool) -> float:
+    """Generate a very made up number representing the Anyness of a type.
+
+    Higher is better, 1.0 is max
+    """
+    t = get_proper_type(ut)
+    if isinstance(t, AnyType) and t.type_of_any != TypeOfAny.special_form:
+        return 0
+    if isinstance(t, NoneType) and arg_pos:
+        return 0.5
+    if isinstance(t, UnionType):
+        if any(isinstance(x, AnyType) for x in t.items):
+            return 0.5
+        if any(has_any_type(x) for x in t.items):
+            return 0.25
+    if isinstance(t, CallableType) and is_tricky_callable(t):
+        return 0.5
+    if has_any_type(t):
+        return 0.5
+
+    return 1.0
+
+
+def any_score_callable(t: CallableType, is_method: bool) -> float:
+    # Ignore the first argument of methods
+    scores = [any_score_type(x, arg_pos=True) for x in t.arg_types[int(is_method):]]
+    # Return type counts twice (since it spreads type information), unless it is
+    # None in which case it does not count at all. (Though it *does* still count
+    # if there are no arguments.)
+    if not isinstance(get_proper_type(t.ret_type), NoneType) or not scores:
+        ret = any_score_type(t.ret_type, arg_pos=False)
+        scores += [ret, ret]
+
+    return sum(scores) / len(scores)
+
+
+def is_tricky_callable(t: CallableType) -> bool:
+    """Is t a callable that we need to put a ... in for syntax reasons?"""
+    return t.is_ellipsis_args or any(
+        k in (ARG_STAR, ARG_STAR2, ARG_NAMED, ARG_NAMED_OPT) for k in t.arg_kinds)
+
 
 class TypeFormatter(TypeStrVisitor):
     """Visitor used to format types
     """
-    # TODO: Generate valid string representation for callable types.
-    # TODO: Probably a bunch more
+    # TODO: Probably a lot
     def __init__(self, module: Optional[str], graph: Graph) -> None:
         super().__init__()
         self.module = module
         self.graph = graph
 
     def visit_instance(self, t: Instance) -> str:
-        if t.type is not None:
-            s = t.type.fullname() or t.type.name() or None
-        else:
-            s = None
+        s = t.type.fullname() or t.type.name() or None
         if s is None:
             return '<???>'
         if s in reverse_builtin_aliases:
@@ -507,7 +639,9 @@ class TypeFormatter(TypeStrVisitor):
         elif t.args != []:
             obj += '[{}]'.format(self.list_str(t.args))
 
-        if mod == 'builtins':
+        if mod_obj == ('builtins', 'unicode'):
+            return 'Text'
+        elif mod == 'builtins':
             return obj
         else:
             delim = '.' if '.' not in obj else ':'
@@ -521,6 +655,37 @@ class TypeFormatter(TypeStrVisitor):
         s = self.list_str(t.items)
         return 'Tuple[{}]'.format(s)
 
+    def visit_union_type(self, t: UnionType) -> str:
+        if len(t.items) == 2 and is_optional(t):
+            return "Optional[{}]".format(remove_optional(t).accept(self))
+        else:
+            return super().visit_union_type(t)
+
+    def visit_callable_type(self, t: CallableType) -> str:
+        # TODO: use extended callables?
+        if is_tricky_callable(t):
+            arg_str = "..."
+        else:
+            # Note: for default arguments, we just assume that they
+            # are required.  This isn't right, but neither is the
+            # other thing, and I suspect this will produce more better
+            # results than falling back to `...`
+            args = [typ.accept(self) for typ in t.arg_types]
+            arg_str = "[{}]".format(", ".join(args))
+
+        return "Callable[{}, {}]".format(arg_str, t.ret_type.accept(self))
+
+
+class StrToText(TypeTranslator):
+    def __init__(self, builtin_type: Callable[[str], Instance]) -> None:
+        self.text_type = builtin_type('builtins.unicode')
+
+    def visit_instance(self, t: Instance) -> Type:
+        if t.type.fullname() == 'builtins.str':
+            return self.text_type
+        else:
+            return super().visit_instance(t)
+
 
 def generate_type_combinations(types: List[Type]) -> List[Type]:
     """Generate possible combinations of a list of types.
@@ -529,7 +694,7 @@ def generate_type_combinations(types: List[Type]) -> List[Type]:
     and unioning the types. We try both.
     """
     joined_type = join_type_list(types)
-    union_type = UnionType.make_simplified_union(types)
+    union_type = make_simplified_union(types)
     if is_same_type(joined_type, union_type):
         return [joined_type]
     else:
@@ -540,31 +705,6 @@ def count_errors(msgs: List[str]) -> int:
     return len([x for x in msgs if ' error: ' in x])
 
 
-def score_type(t: Type) -> int:
-    """Generate a score for a type that we use to pick which type to use.
-
-    Lower is better, prefer non-union/non-any types. Don't penalize optionals.
-    """
-    if isinstance(t, AnyType):
-        return 2
-    if isinstance(t, UnionType):
-        if any(isinstance(x, AnyType) for x in t.items):
-            return 2
-        if not is_optional(t):
-            return 1
-    return 0
-
-
-def score_callable(t: CallableType) -> int:
-    return sum([score_type(x) for x in t.arg_types])
-
-
-def callable_has_any(t: CallableType) -> int:
-    # We count a bare None in argument position as Any, since
-    # pyannotate turns it into Optional[Any]
-    return any(isinstance(at, NoneType) for at in t.arg_types) or has_any_type(t)
-
-
 T = TypeVar('T')
 
 
