@@ -14,7 +14,6 @@ import pytest  # type: ignore  # no pytest in typeshed
 from typing import List, Tuple, Set, Optional, Iterator, Any, Dict, NamedTuple, Union
 
 from mypy.test.config import test_data_prefix, test_temp_dir, PREFIX
-from mypy.test.hacks import new_semanal_blacklist
 
 root_dir = os.path.normpath(PREFIX)
 
@@ -205,8 +204,6 @@ class DataDrivenTestCase(pytest.Item):  # type: ignore  # inheriting from Any
         if ((platform == 'windows' and sys.platform != 'win32')
                 or (platform == 'posix' and sys.platform == 'win32')):
             skip = True
-        if os.getenv('NEWSEMANAL') and os.path.split(file)[-1] in new_semanal_blacklist:
-            skip = True
         self.skip = skip
         self.data = data
         self.line = line
@@ -513,12 +510,14 @@ def expand_errors(input: List[str], output: List[str], fnam: str) -> None:
                 elif m.group(1) == 'W':
                     severity = 'warning'
                 col = m.group('col')
+                message = m.group('message')
+                message = message.replace('\\#', '#')  # adds back escaped # character
                 if col is None:
                     output.append(
-                        '{}:{}: {}: {}'.format(fnam, i + 1, severity, m.group('message')))
+                        '{}:{}: {}: {}'.format(fnam, i + 1, severity, message))
                 else:
                     output.append('{}:{}:{}: {}: {}'.format(
-                        fnam, i + 1, col, severity, m.group('message')))
+                        fnam, i + 1, col, severity, message))
 
 
 def fix_win_path(line: str) -> str:
@@ -567,6 +566,8 @@ def pytest_addoption(parser: Any) -> None:
                     help='Copy the temp directories from failing tests to a target directory')
     group.addoption('--mypy-verbose', action='count',
                     help='Set the verbose flag when creating mypy Options')
+    group.addoption('--mypyc-showc', action='store_true', default=False,
+                    help='Display C code on mypyc test failures')
 
 
 # This function name is special to pytest.  See
