@@ -133,18 +133,23 @@ class FineGrainedSuite(DataSuite):
                 changed = [mod for mod, file in server.fine_grained_manager.changed_modules]
                 targets = server.fine_grained_manager.processed_targets
 
-            assert_module_equivalence(
-                'stale' + str(step - 1),
-                testcase.expected_stale_modules.get(step - 1),
-                changed)
-            assert_module_equivalence(
-                'rechecked' + str(step - 1),
-                testcase.expected_rechecked_modules.get(step - 1),
-                updated)
-            assert_target_equivalence(
-                'targets' + str(step),
-                testcase.expected_fine_grained_targets.get(step),
-                targets)
+            expected_stale = testcase.expected_stale_modules.get(step - 1)
+            if expected_stale is not None:
+                assert_module_equivalence(
+                    'stale' + str(step - 1),
+                    expected_stale, changed)
+
+            expected_rechecked = testcase.expected_rechecked_modules.get(step - 1)
+            if expected_rechecked is not None:
+                assert_module_equivalence(
+                    'rechecked' + str(step - 1),
+                    expected_rechecked, updated)
+
+            expected = testcase.expected_fine_grained_targets.get(step)
+            if expected:
+                assert_target_equivalence(
+                    'targets' + str(step),
+                    expected, targets)
 
             new_messages = normalize_messages(new_messages)
 
@@ -176,6 +181,7 @@ class FineGrainedSuite(DataSuite):
         options.incremental = True
         options.use_builtins_fixtures = True
         options.show_traceback = True
+        options.error_summary = False
         options.fine_grained_incremental = not build_cache
         options.use_fine_grained_cache = self.use_cache and not build_cache
         options.cache_fine_grained = self.use_cache
@@ -191,7 +197,7 @@ class FineGrainedSuite(DataSuite):
         return options
 
     def run_check(self, server: Server, sources: List[BuildSource]) -> List[str]:
-        response = server.check(sources)
+        response = server.check(sources, is_tty=False, terminal_width=-1)
         out = cast(str, response['out'] or response['err'])
         return out.splitlines()
 
@@ -270,9 +276,13 @@ class FineGrainedSuite(DataSuite):
             callsites = '--callsites' in flags
             no_any = '--no-any' in flags
             no_errors = '--no-errors' in flags
+            try_text = '--try-text' in flags
+            m = re.match('--flex-any=([0-9.]+)', flags)
+            flex_any = float(m.group(1)) if m else None
             res = cast(Dict[str, Any],
                        server.cmd_suggest(
                            target.strip(), json=json, no_any=no_any, no_errors=no_errors,
+                           try_text=try_text, flex_any=flex_any,
                            callsites=callsites))
             val = res['error'] if 'error' in res else res['out'] + res['err']
             output.extend(val.strip().split('\n'))
@@ -281,7 +291,7 @@ class FineGrainedSuite(DataSuite):
     def get_suggest(self, program_text: str,
                     incremental_step: int) -> List[Tuple[str, str]]:
         step_bit = '1?' if incremental_step == 1 else str(incremental_step)
-        regex = '# suggest{}: (--[a-zA-Z0-9_\\-./?^ ]+ )*([a-zA-Z0-9_./?^ ]+)$'.format(step_bit)
+        regex = '# suggest{}: (--[a-zA-Z0-9_\\-./=?^ ]+ )*([a-zA-Z0-9_./?^ ]+)$'.format(step_bit)
         m = re.findall(regex, program_text, flags=re.MULTILINE)
         return m
 
